"""
Utilities for generating merged tables of choosers and alternatives, including extensive
sampling functionality.

"""
import numpy as np
import pandas as pd


class MergedChoiceTable(object):
    """
    Generates a merged long-format table of observations (choosers) and alternatives, for 
    discrete choice model estimation or simulation. 
    
    Supports random sampling of alternatives (uniform or weighted). Supports sampling with
    or without replacement. Supports merging observations and alternatives without 
    sampling them. Supports alternative-specific weights, as well as interaction weights
    that depend on both the observation and alternative. Supports automatic merging of 
    interaction terms onto the final data table.
    
    Support is PLANNED for specifying availability of alternatives, specifying random 
    state, and passing interaction-type parameters as callable generator functions.
    
    Does NOT support cases where the number of alternatives in the final table varies 
    across observations.

    Reserved column names: 'chosen'.
    
    Parameters
    ----------
    observations : pandas.DataFrame
        Table with one row for each chooser or choice scenario, with unique ID's in the 
        index field. Additional columns can contain fixed attributes of the choosers. 
        Index name is set to 'obs_id' if none provided. All observation/alternative 
        column names must be unique except for the join key.

    alternatives : pandas.DataFrame
        Table with one row for each alternative, with unique ID's in the index field.
        Additional columns can contain fixed attributes of the alternatives. Index name
        is set to 'alt_id' if none provided. All observation/alternative column names
        must be unique except for the join key.

    chosen_alternatives : str or pandas.Series, optional
        List of the alternative ID selected in each choice scenario. (This is required for
        preparing estimation data, but not for simulation data.) If str, interpreted as a
        column name from the observations table. If Series, it will be joined onto the
        obserations table before processing. The column will be dropped from the merged 
        table and replaced with a binary column named 'chosen'.

    sample_size : int, optional
        Number of alternatives to sample for each choice scenario. If 'None', all of the 
        alternatives will be available for each chooser in the merged table. The sample 
        size includes the chosen alternative, if applicable. If replace=False, the sample
        size must be less than or equal to the total number of alternatives. 

    replace : boolean, optional
        Whether to sample alternatives with or without replacement, at the level of a 
        single chooser or choice scenario. If replace=True (default), alternatives may
        appear multiple times in a single choice set. If replace=False, an alternative
        will appear at most once per choice set. Sampling with replacement is much more
        efficient, so setting replace=False may have performance implications if there are
        very large numbers of observations or alternatives.        
    
    weights : str, pandas.Series, optional
        Numerical weights to apply when sampling alternatives. If str, interpreted as a 
        column from the alternatives table. If Series, it can contain either (a) one 
        weight for each alternative or (b) one weight for each combination of observation 
        and alternative. The former should include a single index with ID's from the 
        alternatives table. The latter should include a MultiIndex with the first level 
        corresponding to the observations table and the second level corresponding to the 
        alternatives table. If callable, it should accept two arguments (obs_id, alt_id) 
        and return the corresponding weight.
        
        TO DO - accept weights specified with respect to derivative characteristics, like
        how the interaction terms work (for example weights could be based on home census
        tract rather than observation id if there are multiple observations per tract)
        
        TO DO - implement support for a callable
    
    availability : pandas.Series or callable, optional (NOT YET IMPLEMENTED)
        Binary representation of the availability of alternatives. Specified and applied 
        similarly to the weights.
    
    interaction_terms : pandas.Series, pandas.DataFrame, or list of either, optional
        Additional column(s) of interaction terms whose values depend on the combination 
        of observation and alternative, to be merged onto the final data table. If passed
        as a Series or DataFrame, it should include a two-level MultiIndex. One level's 
        name and values should match an index or column from the observations table, and 
        the other should match an index or column from the alternatives table. 
        
        TO DO - implement support for a callable
            
    random_state : NOT YET IMPLEMENTED
        Representation of random state, for replicability of the sampling.

    """
    def __init__(self, observations, alternatives, chosen_alternatives=None,
                 sample_size=None, replace=True, weights=None, availability=None,
                 interaction_terms=None, random_state=None):
        
        # Standardize and validate the inputs...
        
        if isinstance(sample_size, float):
            sample_size = int(sample_size)
        
        if (sample_size is not None):
            if (sample_size <= 0):
                raise ValueError("Cannot sample {} alternatives; to run without sampling "
                        "leave sample_size=None".format(sample_size))

            # TO DO - should probably just return as many alternatives as we can (and wait 
            # to evaluate this until after evaluating the sampling filters)
            if (replace == False) & (sample_size > alternatives.shape[0]):
                raise ValueError("Cannot sample without replacement with sample_size {} "
                        "and n_alts {}".format(sample_size, alternatives.shape[0]))
        
        # TO DO - check that dfs have unique indexes
        # TO DO - check that chosen_alternatives correspond correctly to other dfs
        # TO DO - same with weights (could join onto other tables and then split off)
        
        # Normalize chosen_alternatives to a pd.Series
        if (chosen_alternatives is not None) & isinstance(chosen_alternatives, str):
            chosen_alternatives = observations[chosen_alternatives].copy()
            observations = observations.drop(chosen_alternatives.name, axis='columns')
            chosen_alternatives.name = '_' + alternatives.index.name  # avoids conflicts
        
        # Allow missing obs and alts, to support .from_df() constructor     
        if (observations is not None):
        
            # Provide default names for observation and alternatives id's
        
            if (observations.index.name == None):
                observations.index.name = 'obs_id'
        
            if (alternatives.index.name == None):
                alternatives.index.name = 'alt_id'
        
            # Check for duplicate column names
            obs_cols = list(observations.columns) + list(observations.index.names)
            alt_cols = list(alternatives.columns) + list(alternatives.index.names)
            dupes = set(obs_cols) & set(alt_cols)
        
            if len(dupes) > 0:
                raise ValueError("Both input tables contain column {}. Please ensure "
                                 "column names are unique before merging".format(dupes))
        
        # Normalize weights to a pd.Series
        if (weights is not None) & isinstance(weights, str):
            weights = alternatives[weights]
        
        weights_1d = False
        weights_2d = False
        if (weights is not None):
            # TO DO - would be nicer to test using the dimensionality of the index and 
            #   then automatically filter for applicable weights if there are too many
            if (len(weights) == len(alternatives)):
                weights_1d = True
            elif (len(weights) == len(observations) * len(alternatives)):
                weights_2d = True
            else:
                raise ValueError("Length of weights is not aligned with length of "
                                 "alternatives and/or observations")
        
        # TO DO - if user passes a single-column df of weights instead of a series, we
        #   should just silently convert it
        
        self.observations = observations
        self.alternatives = alternatives
        self.chosen_alternatives = chosen_alternatives
        self.sample_size = sample_size
        self.replace = replace
        self.weights = weights
        self.interaction_terms = interaction_terms
        self.random_state = random_state

        self.weights_1d = weights_1d
        self.weights_2d = weights_2d
        
        # Build choice table...
        # Allow missing obs and alts, to support .from_df() constructor     
        if (observations is not None):

            if (len(observations) == 0) or (len(alternatives) == 0):
                self._merged_table = pd.DataFrame()
        
            elif (sample_size is None):
                self._merged_table = self._build_table_without_sampling()
        
            else:
                self._merged_table = self._build_table()
        
        
    @classmethod
    def from_df(cls, df):
        """
        Create a MergedChoiceTable instance from a pre-generated DataFrame.

        Each chooser's rows should be contiguous. If applicable, the chosen alternative
        should be listed first. This ordering is used by MergedChoiceTable.to_frame(),
        and appears to be an undocumented requirement of the legacy MNL code.

        Parameters
        ----------
        df : pandas.DataFrame
            Table with a two-level MultiIndex where the first level corresponds to the
            index of the observations and the second to the index of the alternatives.
            May include a binary column named 'chosen' indicating observed choices.
        
        Returns
        -------
        MergedChoiceTable

        """
        obj = cls(observations = None, alternatives = None)
        obj._merged_table = df
        
        # TO DO: sort the dataframe so that rows are automatically in a consistent order
        
        return obj


    def _merge_interaction_terms(self, df):
        """
        Merges interaction terms (if they exist) onto the input DataFrame. 
        
        Parameters
        ----------
        df : pd.DataFrame
            Should contain two columns whose names match the index levels of the 
            interaction data.
        
        Expected class parameters
        -------------------------
        self.interaction_terms : pd.Series or pd.DataFrame, optional
            Should have a two-level, named MultiIndex. If self.interaction_terms is None, 
            function returns the input DataFrame.
            
        Returns
        -------
        pd.DataFrame
            Same format as input, with interaction term column(s) added.
        
        """
        if (self.interaction_terms is None):
            return df

        if not isinstance(self.interaction_terms, list):
            self.interaction_terms = [self.interaction_terms]

        for intx_table in self.interaction_terms:
            df = df.join(pd.DataFrame(intx_table), how='left', 
                         on=intx_table.index.names)
        
        return df
    
    
    def _build_table_without_sampling(self):
        """
        This handles the cases where each alternative is available for each chooser.
        
        Expected class parameters
        -------------------------
        self.observations : pd.DataFrame
        self.alternatives : pd.DataFrame
        self.chosen_alternatives : pd.Series or None
        self.sample_size : None

        """
        oid_name = self.observations.index.name
        aid_name = self.alternatives.index.name
        
        obs_ids = np.repeat(self.observations.index.values, len(self.alternatives))
        alt_ids = np.tile(self.alternatives.index.values, reps=len(self.observations))
        
        df = pd.DataFrame({oid_name: obs_ids, aid_name: alt_ids})
   
        df = df.join(self.observations, how='left', on=oid_name)
        df = df.join(self.alternatives, how='left', on=aid_name)
        
        if (self.chosen_alternatives is not None):
            df['chosen'] = 0
            df = df.join(self.chosen_alternatives, how='left', on=oid_name)
            df.loc[df[aid_name] == df[self.chosen_alternatives.name], 'chosen'] = 1
            df.drop(self.chosen_alternatives.name, axis='columns', inplace=True)
        
        df = self._merge_interaction_terms(df)
        df.set_index([oid_name, aid_name], inplace=True)
        return df

    
    def _get_availability(self, obs_id):
        """
        Get alternative availability for a single observation id. For now, this just 
        checks whether the chosen alternative is known and if so makes it unavailable.
        
        Parameters
        ----------
        observation_id : value from index of self.observations
        
        Expected class parameters
        -------------------------
        self.alternatives : pd.DataFrame
        self.chosen_alternatives : pd.Series or None
        
        Returns
        -------
        list of booleans
        
        """
        # TO DO - seems inefficient?

        a = np.repeat(True, self.alternatives.shape[0])
        
        if (self.chosen_alternatives is not None):
            a = (self.alternatives.index != self.chosen_alternatives[obs_id])

        return a        
    
    
    def _get_weights(self, obs_id):
        """
        Get sampling weights corresponding to a single observation id.
        
        Parameters
        ----------
        observation_id : value from index of self.observations
        
        Expected class parameters
        -------------------------
        self.weights : pd.Series, callable, or None
        self.chosen_alternatives : pd.Series or None
        self.weights_1d : boolean
        self.weights_2d : boolean
        
        Returns
        -------
        pd.Series of weights
        
        """
        if (self.weights is None):
            return
            
        elif (callable(self.weights)):
            # TO DO - implement
            pass
        
        elif (self.weights_1d == True):
            return self.weights
        
        elif (self.weights_2d == True):
            return self.weights.loc[obs_id]
        
        else:
            raise ValueError  # unexpected inputs
                    

    def _build_table(self):
        """
        Build and return the merged choice table. This handles all cases where sampling
        is performed.
        
        Expected class parameters
        -------------------------
        self.observations : pd.DataFrame
        self.alternatives : pd.DataFrame
        self.chosen_alternatives : pd.Series or None
        self.sample_size : int
        self.replace : boolean
        self.weights : pd.Series, callable, or None
        self.random_state : NOT YET IMPLEMENTED
        
        Returns
        -------
        pd.DataFrame

        """
        n_obs = self.observations.shape[0]
        
        oid_name = self.observations.index.name
        aid_name = self.alternatives.index.name
        
        samp_size = self.sample_size
        if (self.chosen_alternatives is not None):
            samp_size = self.sample_size - 1
        
        obs_ids = np.repeat(self.observations.index.values, samp_size)
        
        # SINGLE SAMPLE: this covers cases where we can draw a single, large sample of 
        # alternatives and distribute them among the choosers, e.g. sampling without 
        # replacement, with optional alternative-specific weights but NOT weights that 
        # apply to combinations of observation x alternative
        
        if (self.replace == True) & (self.weights is None):
            
            alt_ids = np.random.choice(self.alternatives.index.values, 
                                       replace = True,
                                       size = n_obs * samp_size)
        
        elif (self.replace == True) & (self.weights_1d == True):
            
            alt_ids = np.random.choice(self.alternatives.index.values, 
                                       replace = True,
                                       p = self.weights/self.weights.sum(),
                                       size = n_obs * samp_size)
        
        # REPEATED SAMPLES: this covers cases where we have to draw separate samples for
        # each observation, e.g. sampling without replacement, or weights that apply to
        # combinations of observation x alternative
        
        elif (self.replace == False) | (self.weights_2d == True):
            
            alt_ids = []

            for obs_id in self.observations.index.values:
                a = self._get_availability(obs_id)                
                available_alts = self.alternatives.loc[a].index.values
                
                w = self._get_weights(obs_id)
                if (w is not None):
                    w = w.loc[a]/w.loc[a].sum()                
                
                sampled_alts = np.random.choice(available_alts, replace=self.replace, 
                                                p=w, size=samp_size).tolist()
                alt_ids += sampled_alts

        
        # Append chosen ids if applicable
        if (self.chosen_alternatives is not None):
            obs_ids = np.append(obs_ids, self.observations.index.values)
            alt_ids = np.append(alt_ids, self.chosen_alternatives)
            chosen = np.append(np.repeat(0, samp_size * n_obs), np.repeat(1, n_obs))
        
        df = pd.DataFrame({oid_name: obs_ids, aid_name: alt_ids})
   
        df = df.join(self.observations, how='left', on=oid_name)
        df = df.join(self.alternatives, how='left', on=aid_name)
        
        if (self.chosen_alternatives is not None):
            df['chosen'] = chosen
            df.sort_values([oid_name, 'chosen'], ascending=False, inplace=True)
        
        df = self._merge_interaction_terms(df)
        df.set_index([oid_name, aid_name], inplace=True)
        return df
        
    
    def to_frame(self):
        """
        Long-format DataFrame of the merged table. The rows representing alternatives for
        a particular chooser are contiguous, with the chosen alternative listed first if
        applicable. (Unless no sampling is performed, in which case the alternatives are
        listed in order.) The DataFrame includes a two-level MultiIndex. The first level
        corresponds to the index of the observations table and the second to the index of 
        the alternatives table. 
        
        Returns
        -------
        pandas.DataFrame

        """
        return self._merged_table

    
    @property
    def observation_id_col(self):
        """
        Name of column in the merged table containing the observation id. Name and values 
        will match the index of the observations table.
        
        Returns
        -------
        str

        """
        return self._merged_table.index.names[0]

    
    @property
    def alternative_id_col(self):
        """
        Name of column in the merged table containing the alternative id. Name and values
        will match the index of the alternatives table.

        Returns
        -------
        str

        """
        return self._merged_table.index.names[1]

    
    @property
    def choice_col(self):
        """
        Name of the generated column containing a binary representation of whether each
        alternative was chosen in the given choice scenario, if applicable.

        Returns
        -------
        str or None

        """
        if ('chosen' in self._merged_table.columns):
            return 'chosen'
        
        else:
            return None




"""
Utilities for constructing pairwise distance matrices and calculating which
geographies are within various distance bands of some reference geography
"""




import numpy as np
import pandas as pd
from itertools import tee
from scipy.spatial.distance import squareform, pdist




def great_circle_vec(lat1, lng1, lat2, lng2, earth_radius=6371009):
    """
    Vectorized function to calculate the great-circle distance between two
    points or between vectors of points. This function is borrowed from OSMnx:
    https://github.com/gboeing/osmnx

    Parameters
    ----------
    lat1 : float or vector of floats
    lng1 : float or vector of floats
    lat2 : float or vector of floats
    lng2 : float or vector of floats
    earth_radius : numeric
        radius of earth in units in which distance will be returned (default is
        meters)

    Returns
    -------
    distance : float or vector of floats
        distance or vector of distances from (lat1, lng1) to (lat2, lng2) in
        units of earth_radius
    """

    phi1 = np.deg2rad(90 - lat1)
    phi2 = np.deg2rad(90 - lat2)

    theta1 = np.deg2rad(lng1)
    theta2 = np.deg2rad(lng2)

    cos = (np.sin(phi1) * np.sin(phi2) * np.cos(theta1 - theta2) + np.cos(phi1) * np.cos(phi2))
    arc = np.arccos(cos)

    # return distance in units of earth_radius
    distance = arc * earth_radius
    return distance




def great_circle_distance_matrix(df, x, y, earth_radius=6371009,
                                 return_int=True):
    """
    Calculate a pairwise great-circle distance matrix from a DataFrame of
    points. Distances returned are in units of earth_radius (default is meters).

    Parameters
    ----------
    df : pandas DataFrame
        a DataFrame of points, uniquely indexed by place identifier (e.g., tract
        ID or parcel ID), represented by x and y coordinate columns
    x : str
        label of the x coordinate column in the DataFrame
    y : str
        label of the y coordinate column in the DataFrame
    earth_radius : numeric
        radius of earth in units in which distance will be returned (default is
        meters)
    return_int : bool
        if True, convert all distances to integers

    Returns
    -------
    pandas Series
        Multi-indexed distance vector in units of df's values, with top-level
        index representing "from" and second-level index representing "to".
    """

    # calculate pairwise great-circle distances between each row and every
    # other row
    df_dist_matrix = df.apply(lambda row: great_circle_vec(row[y], row[x], df[y], df[x], earth_radius=earth_radius),
                              axis='columns')

    # convert all distances to integers, if so configured
    if return_int:
        df_dist_matrix = df_dist_matrix.fillna(0).astype(int)

    # convert the distance matrix to a multi-indexed vector and return it
    labels = df.index.values
    df_dist_matrix.columns = labels
    df_dist_matrix.index = labels
    return df_dist_matrix.stack()




def euclidean_distance_matrix(df):
    """
    Calculate a pairwise euclidean distance matrix from a DataFrame of points.
    Distances returned are in units of x and y columns.

    Parameters
    ----------
    df : pandas DataFrame
        a DataFrame of points, uniquely indexed by place identifier (e.g., tract
        ID or parcel ID), represented by x and y coordinate columns

    Returns
    -------
    pandas Series
        Multi-indexed distance vector in units of df's values, with top-level
        index representing "from" and second-level index representing "to".
    """

    # calculate pairwise euclidean distances between all rows
    distance_vector = pdist(X=df, metric='euclidean')

    # convert the distance vector to a square distance matrix
    dist_matrix = squareform(distance_vector)

    # convert the distance matrix to a multi-indexed vector and return it
    labels = df.index.values
    df_dist_matrix = pd.DataFrame(data=dist_matrix, columns=labels, index=labels)
    return df_dist_matrix.stack()




def network_distance_matrix(df, x, y):
    """
    Calculate a pairwise network distance matrix from a DataFrame of points.

    Parameters
    ----------
    df : pandas DataFrame
        a DataFrame of points, uniquely indexed by place identifier (e.g., tract
        ID or parcel ID), represented by x and y coordinate columns
    x : str
        label of the x coordinate column in the DataFrame
    y : str
        label of the y coordinate column in the DataFrame

    Returns
    -------
    None
    """

    # placeholder until implementation of network distance matrix calculation
    raise NotImplementedError('network_distance_matrix is not currently implemented')




def distance_matrix(df, method='euclidean', x='lng', y='lat',
                    earth_radius=6371009, return_int=True):
    """
    Calculate a pairwise distance matrix from a DataFrame of two-dimensional
    points.

    Parameters
    ----------
    df : pandas DataFrame
        a DataFrame of points, uniquely indexed by place identifier (e.g., tract
        ID or parcel ID), represented by x and y coordinate columns
    method : str
        {'euclidean', 'greatcircle', 'network'}
        which algorithm to use for calculating pairwise distances
    x : str
        if method='greatcircle' or 'network', label of the x coordinate column
        in the DataFrame
    y : str
        if method='greatcircle' or 'network', label of the y coordinate column
        in the DataFrame
    earth_radius : numeric
        if method='greatcircle', radius of earth in units in which distance will
        be returned (default is meters)
    return_int : bool
        if method='greatcircle', if True, convert all distances to integers

    Returns
    -------
    pandas Series
        Multi-indexed distance vector in units of df's values, with top-level
        index representing "from" and second-level index representing "to".
    """

    if not df.index.is_unique:
        raise ValueError('The passed-in DataFrame must have a unique index')

    if method == 'euclidean':
        return euclidean_distance_matrix(df=df)
    elif method == 'greatcircle':
        return great_circle_distance_matrix(df=df, x=x, y=y,
                                            earth_radius=earth_radius,
                                            return_int=return_int)
    elif method == 'network':
        return network_distance_matrix(df=df, x=x, y=y)
    else:
        raise ValueError('argument `method` must be one of "euclidean", "greatcircle", or "network"')




def pairwise(iterable):
    """
    Iterate through a list, pairwise.

    Parameters
    ----------
    iterable : list-like
        the list-like object to iterate through pairwise

    Returns
    -------
    zip
        a zipped iterable of pairwise tuples
    """

    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)




def distance_bands(dist_vector, distances):
    """
    Identify all geographies located within each distance band of each
    geography.

    The list of distances is treated pairwise to create distance bands, with
    the first element of each pair forming the band's inclusive lower limit and
    the second element of each pair forming the band's exclusive upper limit.
    For example, if distances=[0, 10, 30], band 0 will contain all geographies
    with a distance >= 0 and < 10 units (e.g., meters) from the reference
    geography, and band 1 will contain all geographies with a distance >= 10
    and < 30 units from the reference geography.

    To make the final distance band include all geographies beyond a certain
    distance, make the final value in the distances list np.inf.

    Parameters
    ----------
    dist_vector : pandas Series
        Multi-indexed distance vector in units of df's values, with top-level
        index representing "from" and second-level index representing "to".
    distances : list
        a list of distance band increments

    Returns
    -------
    pandas Series
        a series multi-indexed by geography ID and distance band number, with
        values of arrays of geography IDs with the corresponding distances from
        that ID
    """

    # loop through each row in distance matrix, identifying all geographies
    # within each distance band of the row's geography
    bands = {}
    dist_matrix = dist_vector.unstack()
    for _, row in dist_matrix.iterrows():
        bands[row.name] = {}

        # for each distance band
        for band_number, (dist1, dist2) in enumerate(pairwise(distances)):

            # find all the geographies within this distance band
            mask = (row >= dist1) & (row < dist2)
            place_ids = row[mask].index.values

            # store value as array of geography IDs keyed by reference geography
            # ID and distance band number
            bands[row.name][band_number] = place_ids

    # convert geography bands to a series multi-indexed by geography ID and
    # distance band number
    return pd.DataFrame(bands).T.stack()



"""
Utilities for Monte Carlo simulation of choices.

"""
import numpy as np
import pandas as pd
from multiprocessing import Process, Manager, Array, cpu_count
from tqdm import tqdm
import warnings


def monte_carlo_choices(probabilities):
    """
    Monte Carlo simulation of choices for a set of K scenarios, each having different
    probability distributions (and potentially different alternatives). 
    
    Choices are independent and unconstrained, meaning that the same alternative can be 
    chosen in multiple scenarios.
    
    This function is equivalent to applying np.random.choice() to each of the K scenarios,
    but it's implemented as a single-pass matrix calculation. When the number of scenarios 
    is large, this is about 50x faster than using df.apply() or a loop.
    
    If all the choice scenarios have the same probability distribution among alternatives,
    you don't need this function. You can use np.random.choice() with size=K, which will 
    be more efficient. (For example, that would work for a choice model whose expression 
    includes only attributes of the alternatives.)

    NOTE ABOUT THE INPUT FORMATS: It's important for the probabilities to be structured
    correctly. This is computationally expensive to verify, so you will not get a warning
    if it's wrong! (TO DO: we should provide an option to perform these checks, though)
    
    1. Probabilities (pd.Series) must include a two-level MultiIndex, the first level 
       representing the scenario (observation) id and the second the alternative id.

    2. Probabilities must be sorted so that each scenario's alternatives are consecutive.
    
    3. Each scenario must have the same number of alternatives. You can pad a scenario 
       with zero-probability alternatives if needed.
       
    4. Each scenario's alternative probabilities must sum to 1. 
    
    Parameters
    ----------
    probabilities: pd.Series
        List of probabilities for each observation (choice scenario) and alternative. 
        Please verify that the formatting matches the four requirements described above.
    
    Returns
    -------
    pd.Series
        List of chosen alternative id's, indexed with the observation id.
    
    """
    # TO DO - if input is a single-column dataframe, silently convert it to series

    obs_name, alts_name = probabilities.index.names

    obs = probabilities.index.get_level_values(0)
    alts = probabilities.index.get_level_values(1)
    
    num_obs = obs.unique().size
    num_alts = probabilities.size // num_obs
    
    # This Monte Carlo approach is adapted from urbansim.urbanchoice.mnl_simulate()
    probs = np.array(probabilities)
    cumprobs = probs.reshape((num_obs, num_alts)).cumsum(axis=1)
    
    # Simulate choice by subtracting a random float
    scaledprobs = np.subtract(cumprobs, np.random.rand(num_obs, 1))

    # Replace negative values with 0 and positive values with 1, then use argmax to
    # return the position of the first postive value
    choice_ix = np.argmax((scaledprobs + 1.0).astype('i4'), axis=1)
    choice_ix_1d = choice_ix + (np.arange(num_obs) * num_alts)
    
    choices = pd.DataFrame({obs_name: obs.values.take(choice_ix_1d),
                            alts_name: alts.values.take(choice_ix_1d)})
    
    return choices.set_index(obs_name)[alts_name]


def iterative_lottery_choices(choosers, alternatives, mct_callable, probs_callable, 
        alt_capacity=None, chooser_size=None, max_iter=None, chooser_batch_size=None):
    """
    Monte Carlo simulation of choices for a set of choice scenarios where (a) the 
    alternatives have limited capacity and (b) the choosers have varying probability 
    distributions over the alternatives. 
    
    Effectively, we simulate the choices sequentially, each time removing the chosen
    alternative or reducing its available capacity. (It's actually done in batches for
    better performance, but the outcome is equivalent.) This requires sampling 
    alternatives and calculating choice probabilities multiple times, which is why
    callables for those actions are required inputs.
    
    Chooser priority is randomized. Capacities can be specified as counts (number of 
    choosers that can be accommodated) or as amounts (e.g. square footage) with 
    corresponding chooser sizes. If total capacity is insufficient to accommodate all the 
    choosers, as many choices will be simulated as possible.
    
    Note that if all the choosers are the same size and have the same probability 
    distribution over alternatives, you don't need this function. You can use
    np.random.choice() with size=K to draw chosen alternatives, which will be more 
    efficient. (This function also works, though.)
    
    Parameters
    ----------
    choosers : pd.DataFrame
        Table with one row for each chooser or choice scenario, with unique ID's in the
        index field. Additional columns can contain fixed attributes of the choosers.
        (Reserved column names: '_size'.)
    
    alternatives : pd.DataFrame
        Table with one row for each alternative, with unique ID's in the index field.
        Additional columns can contain fixed attributes of the alternatives. (Reserved 
        column names: '_capacity'.)
    
    mct_callable : callable
        Callable that samples alternatives to generate a table of choice scenarios. It 
        should accept subsets of the choosers and alternatives tables and return a 
        choicemodels.tools.MergedChoiceTable.
    
    probs_callable : callable
        Callable that generates predicted probabilities for a table of choice scenarios.
        It should accept a choicemodels.tools.MergedChoiceTable and return a pd.Series
        with indexes matching the input.
    
    alt_capacity : str, optional
        Name of a column in the alternatives table that expresses the capacity of 
        alternatives. If not provided, each alternative is interpreted as accommodating a
        single chooser.
    
    chooser_size : str, optional
        Name of a column in the choosers table that expresses the size of choosers. 
        Choosers might have varying sizes if the alternative capacities are amounts 
        rather than counts -- e.g. square footage or employment capacity. Chooser sizes 
        must be in the same units as alternative capacities. If not provided, each chooser
        has a size of 1. 
    
    max_iter : int or None, optional
        Maximum number of iterations. If None (default), the algorithm will iterate until 
        all choosers are matched or no alternatives remain.

    chooser_batch_size : int or None, optional
        Size of the batches for processing smaller groups of choosers one at a time. 
        Useful when the anticipated size of the merged choice tables (choosers X 
        alternatives X covariates) will be too large for python/pandas to handle. 

    Returns
    -------
    pd.Series
        List of chosen alternative id's, indexed with the chooser (observation) id. 
            
    """
    # TO DO - how does MCT handle sample size greater than the number of available alts?
    
    # Copy the input dataframes to avoid editing the originals
    choosers = choosers.copy()
    alts = alternatives.copy()
    
    if alt_capacity is None:
        alt_capacity = '_capacity'
        alts.loc[:,alt_capacity] = 1
    
    if chooser_size is None:
        chooser_size = '_size'
        choosers.loc[:,chooser_size] = 1
    
    capacity, size = (alt_capacity, chooser_size)
    
    len_choosers = len(choosers)
    valid_choices = pd.Series()
    iter = 0
    
    while (len(valid_choices) < len_choosers):
        iter += 1
        if max_iter is not None:
            if (iter > max_iter):
                break
        if alts[capacity].max() < choosers[size].min():
            print("{} choosers cannot be allocated.".format(len(choosers)))
            print("\nRemaining capacity on alternatives but not enough to accodomodate choosers' sizes")
            break
        if chooser_batch_size is None or chooser_batch_size > len(choosers):
            mct = mct_callable(choosers.sample(frac=1), alts)
        else:
            mct = mct_callable(choosers.sample(chooser_batch_size), alts)

        if len(mct.to_frame()) == 0:
            print("No valid alternatives for the remaining choosers")
            break

        probs = probs_callable(mct)
        choices = pd.DataFrame(monte_carlo_choices(probs))

        # join capacities and sizes
        oid, aid = (mct.observation_id_col, mct.alternative_id_col)
        c = choices.join(alts[capacity], on=aid).join(choosers[size], on=oid)
        c.loc[:,'_cumsize'] = c.groupby(aid)[size].cumsum()

        # save valid choices
        c_valid = (c._cumsize <= c[capacity])
        valid_choices = pd.concat([valid_choices, c[aid].loc[c_valid]])

        print("Iteration {}: {} of {} valid choices".format(iter, len(valid_choices), 
                len_choosers))

        # update choosers and alternatives
        choosers = choosers.drop(c.loc[c_valid].index.values)
        # print("{} remaining choosers".format(len(choosers)))

        placed_capacity = c.loc[c_valid].groupby(aid)._cumsize.max()
        alts.loc[:,capacity] = alts[capacity].subtract(placed_capacity, fill_value=0)

        full = alts.loc[alts[capacity] == 0]
        alts = alts.drop(full.index)
        # print("{} remaining alternatives".format(len(alts)))

    # retain original index names
    valid_choices.index.name = choosers.index.name
    valid_choices.name = alts.index.name

    return valid_choices


def _parallel_lottery_choices_worker(
        choosers, alternatives, choices_dict, chosen_alts,
        mct_callable, probs_callable, alt_capacity=None,
        chooser_size=None, proc_num=0, batch_size=0):  # pragma: no cover

    """
    Worker process called only by the parallel_lottery_choices() method.

    Parameters
    ----------
    choosers : pd.DataFrame
        Table with one row for each chooser or choice scenario, with unique ID's in the
        index field. Additional columns can contain fixed attributes of the choosers.
        (Reserved column names: '_size'.)
    
    alternatives : pd.DataFrame
        Table with one row for each alternative, with unique ID's in the index field.
        Additional columns can contain fixed attributes of the alternatives. (Reserved 
        column names: '_capacity'.)

    choices_dict : multiprocessing.managers.SyncManager.dict
        A dictionary array allocated from shared memory

    chosen_alts : multiprocessing.Array
        A ctypes array allocated from shared memory
    
    mct_callable : callable
        Callable that samples alternatives to generate a table of choice scenarios. It 
        should accept subsets of the choosers and alternatives tables and return a 
        choicemodels.tools.MergedChoiceTable.
    
    probs_callable : callable
        Callable that generates predicted probabilities for a table of choice scenarios.
        It should accept a choicemodels.tools.MergedChoiceTable and return a pd.Series
        with indexes matching the input.
    
    alt_capacity : str, optional
        Name of a column in the alternatives table that expresses the capacity of 
        alternatives. If not provided, each alternative is interpreted as accommodating a
        single chooser.
    
    chooser_size : str, optional
        Name of a column in the choosers table that expresses the size of choosers. 
        Choosers might have varying sizes if the alternative capacities are amounts 
        rather than counts -- e.g. square footage or employment capacity. Chooser sizes 
        must be in the same units as alternative capacities. If not provided, each chooser
        has a size of 1. 
    
    proc_num : int
        Integer representing the sequential order in which the worker was spawned

    batch_size : int
        Integer representing the chooser batch size

    """

    st_choice_idx = proc_num * batch_size
    if alt_capacity is None:
        alt_capacity = '_capacity'
    if chooser_size is None:
        chooser_size = '_size'
        choosers.loc[:, chooser_size] = 1

    capacity, size = (alt_capacity, chooser_size)

    len_choosers = len(choosers)
    alts_name = alternatives.index.name
    valid_choices = pd.Series()
    max_mct_size = 0

    iter = 0
    while (len(valid_choices) < len_choosers):
        chosen_alts_list = list(chosen_alts.get_obj())
        alternatives = alternatives[~alternatives.index.isin(chosen_alts_list)]
        iter += 1

        if alternatives['_capacity'].max() < choosers[size].min():
            print("{} choosers cannot be allocated.".format(len(choosers)))
            print("\nRemaining capacity on alternatives but "
                  "not enough to accodomodate choosers' sizes")
            break

        mct = mct_callable(choosers.sample(frac=1), alternatives)
        mct_size = len(mct.to_frame())
        max_mct_size = max(max_mct_size, mct_size)
        if mct_size == 0:
            # print("No valid alternatives for the remaining choosers")
            break

        probs = probs_callable(mct)
        choices = pd.DataFrame(monte_carlo_choices(probs))

        # join capacities and sizes
        oid, aid = (mct.observation_id_col, mct.alternative_id_col)
        c = choices.join(
            alternatives[capacity], on=aid).join(
            choosers[size], on=oid)

        # when size==1, _cumsize counts the cumulative number of times
        # each alternative appears. thus, below, c_valid creates a
        # mask that retains just the choice of the chooser who chose
        # their alternative first, provided that choice hasn't been
        # made elsewhere as documented by the shared chosen_alts list
        c.loc[:, '_cumsize'] = c.groupby(aid)[size].cumsum()

        # the shared array of chosen alts must stay locked by the
        # current worker between the time the worker converts it
        # to a list to make sure the workers choices haven't been
        # chosen already and the time the worker updates the array
        with chosen_alts.get_lock():

            chosen_alts_list = list(chosen_alts.get_obj())
            c_valid = (c._cumsize <= c[capacity]) & (
                ~c[alts_name].isin(chosen_alts_list))
            iter_valid_choices = c[aid].loc[c_valid]
            if len(iter_valid_choices) == 0:
                continue

            num_valid_choices = len(iter_valid_choices.values)
            chosen_alts[st_choice_idx:st_choice_idx + num_valid_choices] = \
                iter_valid_choices.values

        st_choice_idx += num_valid_choices
        alternatives.drop(iter_valid_choices.values, inplace=True)

        iter_valid_choices.index.name = choosers.index.name
        iter_valid_choices.name = alternatives.index.name
        choices_dict.update(iter_valid_choices.to_dict())
        valid_choices = pd.concat([valid_choices, iter_valid_choices])

        choosers = choosers.drop(c.loc[c_valid].index.values)
    return


def parallel_lottery_choices(
        choosers, alternatives, mct_callable, probs_callable,
        alt_capacity=None, chooser_size=None, chooser_batch_size=None):
    """
    A parallelized version of the iterative_lottery_choices method. Chooser
    batches are processed in parallel rather than sequentially.

    NOTE: In it's current form, this method is only supported for simulating
    choices where every alternative has a capacity of 1. 

    Parameters
    ----------
    choosers : pd.DataFrame
        Table with one row for each chooser or choice scenario, with unique ID's in the
        index field. Additional columns can contain fixed attributes of the choosers.
        (Reserved column names: '_size'.)
    
    alternatives : pd.DataFrame
        Table with one row for each alternative, with unique ID's in the index field.
        Additional columns can contain fixed attributes of the alternatives. (Reserved 
        column names: '_capacity'.)
    
    mct_callable : callable
        Callable that samples alternatives to generate a table of choice scenarios. It 
        should accept subsets of the choosers and alternatives tables and return a 
        choicemodels.tools.MergedChoiceTable.
    
    probs_callable : callable
        Callable that generates predicted probabilities for a table of choice scenarios.
        It should accept a choicemodels.tools.MergedChoiceTable and return a pd.Series
        with indexes matching the input.
    
    alt_capacity : str, optional
        Name of a column in the alternatives table that expresses the capacity of 
        alternatives. If not provided, each alternative is interpreted as accommodating a
        single chooser.
    
    chooser_size : str, optional
        Name of a column in the choosers table that expresses the size of choosers. 
        Choosers might have varying sizes if the alternative capacities are amounts 
        rather than counts -- e.g. square footage or employment capacity. Chooser sizes 
        must be in the same units as alternative capacities. If not provided, each chooser
        has a size of 1. 
    
    max_iter : int or None, optional
        Maximum number of iterations. If None (default), the algorithm will iterate until 
        all choosers are matched or no alternatives remain.

    chooser_batch_size : int or None, optional
        Size of the batches for processing smaller groups of choosers one at a time. Useful
        when the anticipated size of the merged choice tables (choosers X alternatives
        X covariates) will be too large for python/pandas to handle.


    Returns
    -------
    pd.Series
        List of chosen alternative id's, indexed with the chooser (observation) id. 

    """
    
    choosers = choosers.copy()
    alternatives = alternatives.copy()

    if alt_capacity is None:
        alt_capacity = '_capacity'
        alternatives.loc[:, alt_capacity] = 1

    if chooser_size is None:
        chooser_size = '_size'
        choosers.loc[:, chooser_size] = 1

    if chooser_batch_size is None or chooser_batch_size > len(choosers):
        obs_batches = [choosers.index.values]
    else:
        obs_batches = [
            choosers.index.values[x:x + chooser_batch_size] for
            x in range(0, len(choosers), chooser_batch_size)]
        num_cpus = cpu_count()
        num_batches = len(obs_batches)
        if num_batches > num_cpus:
            warnings.warn(
                "The specified batch size yields more batches than there "
                "are vCPU's for parallel processing on this computer "
                "({0} vs. {1}). To optimize this code, choose a larger "
                "batch size or consider using the iterative_lottery_choices() "
                "method instead.".format(num_batches, num_cpus))

    manager = Manager()
    shared_choices_dict = manager.dict()
    alternatives[alt_capacity] = 1
    shared_chosen_alts = Array('i', len(choosers))
    jobs = []
    for b, batch in enumerate(obs_batches):
        obs = choosers.loc[batch]
        proc = Process(
            target=_parallel_lottery_choices_worker,
            args=(
                obs, alternatives, shared_choices_dict, shared_chosen_alts,
                mct_callable, probs_callable, alt_capacity, chooser_size,
                b, chooser_batch_size)
        )
        proc.start()
        jobs.append(proc)

    for j, job in tqdm(enumerate(jobs), total=len(jobs)):
        job.join()

    choices_dict = shared_choices_dict._getvalue()

    # convert choices dict to series with original index names
    out_choices = pd.Series(choices_dict)
    out_choices.index.name = choosers.index.name
    out_choices.name = alternatives.index.name

    return out_choices



from __future__ import print_function

import datetime
import logging
from collections import OrderedDict

import numpy as np
import pandas as pd
import pylogit
import scipy.optimize
import scipy.stats
from patsy import dmatrix
from statsmodels.iolib.table import SimpleTable

from .tools import MergedChoiceTable
from .tools import pmat
from .tools.pmat import PMAT


"""
#####################
NEW CLASS DEFINITIONS
#####################

"""


class MultinomialLogit(object):
    """
    A class with methods for estimating multinomial logit discrete choice models. Each
    observation is a choice scenario in which a chooser selects one alternative from a
    choice set of two or more. The fitted parameters represent a joint optimization of
    utility expressions that explains observed choices based on attributes of the
    alternatives and of the choosers.

    The input data needs to be in "long" format, with one row for each combination of
    chooser and alternative. Columns contain relevant attributes and identifiers. (If the
    choice sets are large, sampling of alternatives should be carried out before data is
    passed to this class.)

    The class constructor supports two use cases:

    1. The first use case is simpler and requires fewer inputs. Each choice scenario must
       have the same number of alternatives, and each alternative must have the same model
       expression (utility equation). This is typical when the alternatives are relatively
       numerous and homogenous, for example with travel destination choice or household
       location choice.

       The following parameters are required: 'data', 'observation_id_col', 'choice_col',
       'model_expression' in Patsy format. If data is provided as a MergedChoiceTable,
       the observation id and choice column names can be read directly from its metadata.

       To fit this type of model, ChoiceModels will use its own estimation engine adapted
       from the UrbanSim MNL codebase.

       Migration from 'urbansim.urbanchoice': Note that these requirements differ from
       the old UrbanSim codebase in a couple of ways. (1) The chosen alternatives need to
       be indicated in a column of the estimation data table instead of in a separate
       matrix, and (2) in lieu of indicating the number of alternatives in each choice
       set, the estimation data table should include an observation id column. These
       changes make the API more consistent with other use cases. See the
       MergedChoiceTable() class for tools and code examples to help with migration.

    2. The second use case is more flexible. Choice scenarios can have varying numbers of
       alternatives, and the model expression (utility equation) can be different for
       distinct alternatives. This is typical when there is a small number of alternatives
       whose salient characteristics vary, for example with travel mode choice.

       The following parameters are required: 'data', 'observation_id_col',
       'alternative_id_col', 'choice_col', 'model_expression' in PyLogit format,
       'model_labels' in PyLogit format (optional).

       To fit this type of model, ChoiceModels will use the PyLogit estimation engine.

    With either use case, the model expression can include attributes of both the choosers
    and the alternatives. Attributes of a particular alternative may vary for different
    choosers (distance, for example), but this must be set up manually in the input data.

    Note that prediction methods are in a separate class: see MultinomialLogitResults().

    Parameters
    ----------

    data : pd.DataFrame or choicemodels.tools.MergedChoiceTable
        A table of estimation data in "long" format, with one row for each combination of
        chooser and alternative. Column labeling must be consistent with the
        'model_expression'. May include extra columns.

    model_expression : Patsy 'formula-like' or PyLogit 'specification'
        For the simpler use case where each choice scenario has the same number of
        alternatives and each alternative has the same model expression, this should be a
        Patsy formula representing the right-hand side of the single model expression.
        This can be a string or a number of other data types. See here:
        https://patsy.readthedocs.io/en/v0.1.0/API-reference.html#patsy.dmatrix

        For the more flexible use case where choice scenarios have varying numbers of
        alternatives or the model expessions vary, this should be a PyLogit OrderedDict
        model specification. See here:
        https://github.com/timothyb0912/pylogit/blob/master/pylogit/pylogit.py#L116-L130

    observation_id_col : str, optional
        Name of column or index containing the observation id. This should uniquely 
        identify each distinct choice scenario. Not required if data is passed as a 
        MergedChoiceTable.

    choice_col : str, optional
        Name of column containing an indication of which alternative has been chosen in
        each scenario. Values should evaluate as binary: 1/0, True/False, etc. Not
        required if data is passed as a MergedChoiceTable.

    model_labels : PyLogit 'names', optional
        If the model expression is a PyLogit OrderedDict, you can provide a corresponding
        OrderedDict of labels. See here:
        https://github.com/timothyb0912/pylogit/blob/master/pylogit/pylogit.py#L151-L165

    alternative_id_col : str, optional
        Name of column or index containing the alternative id. This is only required if 
        the model expression varies for different alternatives. Not required if data is 
        passed as a MergedChoiceTable.

    initial_coefs : numeric or list-like of numerics, optional
        Initial coefficients (beta values) to begin the optimization process with. Provide
        a single value for all coefficients, or an array containing a value for each
        one being estimated. If None, initial coefficients will be 0.

    weights : 1D array, optional
        NOT YET IMPLEMENTED - Estimation weights.

    """
    def __init__(self, data, model_expression, observation_id_col=None, choice_col=None,
                 model_labels=None, alternative_id_col=None, initial_coefs=None,
                 weights=None):
        self._data = data
        self._model_expression = model_expression
        self._observation_id_col = observation_id_col
        self._alternative_id_col = alternative_id_col
        self._choice_col = choice_col
        self._model_labels = model_labels
        self._initial_coefs = initial_coefs
        self._weights = weights

        if isinstance(self._data, MergedChoiceTable):
            self._df = self._data.to_frame()
            self._observation_id_col = self._data.observation_id_col
            self._alternative_id_col = self._data.alternative_id_col
            self._choice_col = self._data.choice_col
        
        else:
            self._df = self._data
        
        if isinstance(self._model_expression, OrderedDict):
            self._estimation_engine = 'PyLogit'

            # parse initial_coefs
            if isinstance(self._initial_coefs, np.ndarray):
                pass
            elif isinstance(self._initial_coefs, list):
                self._initial_coefs = np.array(self._initial_coefs)
            elif (self._initial_coefs == None):
                self._initial_coefs = np.zeros(len(self._model_expression))
            else:
                self._initial_coefs = np.repeat(self._initial_coefs,
                                                len(self._model_expression))

        else:
            self._estimation_engine = 'ChoiceModels'
            self._numobs = self._df.reset_index()[[self._observation_id_col]].\
                                    drop_duplicates().shape[0]
            self._numalts = self._df.shape[0] // self._numobs

        return

    
    def fit(self):
        """
        Fit the model using maximum likelihood estimation. Uses either the ChoiceModels
        or PyLogit estimation engine as appropriate.

        Returns
        -------
        MultinomialLogitResults() object.

        """
        if (self._estimation_engine == 'PyLogit'):

            m = pylogit.create_choice_model(data = self._df,
                                            obs_id_col = self._observation_id_col,
                                            alt_id_col = self._alternative_id_col,
                                            choice_col = self._choice_col,
                                            specification = self._model_expression,
                                            names = self._model_labels,
                                            model_type = 'MNL')

            m.fit_mle(init_vals = self._initial_coefs)
            results = MultinomialLogitResults(estimation_engine = self._estimation_engine,
                                              model_expression = self._model_expression,
                                              results = m)

        elif (self._estimation_engine == 'ChoiceModels'):

            dm = dmatrix(self._model_expression, data=self._df)

            chosen = np.reshape(self._df[[self._choice_col]].values,
                                (self._numobs, self._numalts))

            log_lik, fit = mnl_estimate(np.array(dm), chosen, self._numalts)

            result_params = dict(log_likelihood = log_lik,
                                 fit_parameters = fit,
                                 x_names = dm.design_info.column_names)

            results = MultinomialLogitResults(estimation_engine = self._estimation_engine,
                                              model_expression = self._model_expression,
                                              results = result_params)

        return results


    @property
    def estimation_engine(self):
        """
        'ChoiceModels' or 'PyLogit'.

        """
        return self._estimation_engine


class MultinomialLogitResults(object):
    """
    The results class represents a fitted model. It can report the model fit, generate
    choice probabilties, etc.
    
    A full-featured results object is returned by MultinomialLogit.fit(). A results object
    with more limited functionality can also be built directly from fitted parameters and
    a model expression.

    Parameters
    ----------
    model_expression : str or OrderedDict
        Patsy 'formula-like' (str) or PyLogit 'specification' (OrderedDict).
    
    results : dict or object, optional
        Raw results as currently provided by the estimation engine. This should be
        replaced with a more consistent and comprehensive set of inputs.

    fitted_parameters : list of floats, optional
        If not provided, these will be extracted from the raw results.

    estimation_engine : str, optional
        'ChoiceModels' (default) or 'PyLogit'.

    """
    def __init__(self, model_expression, results=None, fitted_parameters=None, 
                 estimation_engine='ChoiceModels'):
        
        if (fitted_parameters is None) & (results is not None):
            if (estimation_engine == 'ChoiceModels'):
                fitted_parameters = results['fit_parameters']['Coefficient'].tolist()

        self.estimation_engine = estimation_engine
        self.model_expression = model_expression
        self.results = results
        self.fitted_parameters = fitted_parameters
        
    
    def __repr__(self):
    	return self.report_fit()

    
    def __str__(self):
        return self.report_fit()

    
    def get_raw_results(self):
        """
        Return the raw results as provided by the estimation engine. Dict or object.

        """
        return self.results

    
    def probabilities(self, data):
        """
        Generate predicted probabilities for a table of choice scenarios, using the fitted
        parameters stored in the results object.
        
        Parameters
        ----------
        data : choicemodels.tools.MergedChoiceTable
            Long-format table of choice scenarios. TO DO - accept other data formats.
        
        Expected class parameters
        -------------------------
        self.model_expression : patsy string
        self.fitted_parameters : list of floats
        
        Returns
        -------
        pandas.Series with indexes matching the input
        
        """
        # TO DO - make sure this handles pylogit case
        
        # TO DO - does MergedChoiceTable guarantee that alternatives for a single scenario
        # are consecutive? seems like a requirement here; should document it
        
        df = data.to_frame()
        numalts = data.sample_size  # TO DO - make this an official MCT param
        
        dm = dmatrix(self.model_expression, data=df)
        
        # utility is sum of data values times fitted betas
        u = np.dot(self.fitted_parameters, np.transpose(dm))
        
        # reshape so axis 0 lists alternatives and axis 1 lists choosers
        u = np.reshape(u, (numalts, u.size // numalts), order='F')
    
        # scale the utilities to make exponentiation easier
        # https://stats.stackexchange.com/questions/304758/softmax-overflow
        u = u - u.max(axis=0)
        
        exponentiated_utility = np.exp(u)
        sum_exponentiated_utility = np.sum(exponentiated_utility, axis=0)
        
        probs = exponentiated_utility / sum_exponentiated_utility
        
        # convert back to ordering of the input data
        probs = probs.flatten(order='F')
        
        df['prob'] = probs  # adds indexes
        return df.prob
    
    
    def report_fit(self):
        """
        Print a report of the model estimation results.

        """
        if (self.estimation_engine == 'PyLogit'):
            output = self.results.get_statsmodels_summary().as_text()

        elif (self.estimation_engine == 'ChoiceModels'):

            # Pull out individual results components
            ll = self.results['log_likelihood']['convergence']
            ll_null = self.results['log_likelihood']['null']

            rho_bar_squared = self.results['log_likelihood']['rho_bar_squared']
            rho_squared = self.results['log_likelihood']['rho_squared']
            df_model = self.results['log_likelihood']['df_model']
            df_resid = self.results['log_likelihood']['df_resid']
            num_obs = self.results['log_likelihood']['num_obs']
            bic = self.results['log_likelihood']['bic']
            aic = self.results['log_likelihood']['aic']

            x_names = self.results['x_names']
            coefs = self.results['fit_parameters']['Coefficient'].tolist()
            std_errs = self.results['fit_parameters']['Std. Error'].tolist()
            t_scores = self.results['fit_parameters']['T-Score'].tolist()
            p_values = self.results['fit_parameters']['P-Values'].tolist()

            def time_now(*args, **kwds):
                now = datetime.datetime.now()
                return now.strftime('%H:%M')

            def date_now(*args, **kwds):
                now = datetime.datetime.now()
                return now.strftime('%Y-%m-%d')

            (header, body) = summary_table(dep_var = 'chosen',
                                           model_name = 'Multinomial Logit',
                                           method = 'Maximum Likelihood',
                                           log_likelihood = ll,
                                           null_log_likelihood = ll_null,
                                           rho_squared = rho_squared,
                                           rho_bar_squared = rho_bar_squared,
                                           df_model = df_model,
                                           df_resid = df_resid,
                                           x_names = x_names,
                                           coefs = coefs,
                                           std_errs = std_errs,
                                           t_scores = t_scores,
                                           p_values = p_values,
                                           bic = bic,
                                           aic = aic,
                                           num_obs = num_obs,
                                           time = time_now(),
                                           date = date_now())

            output = header.as_text() + '\n' + body.as_text()

        return output


def summary_table(title=None, dep_var='', model_name='', method='', date='',
                  time='', aic=None, bic=None, num_obs=None, df_resid=None,
                  df_model=None, rho_squared=None, rho_bar_squared=None,
                  log_likelihood=None, null_log_likelihood=None, x_names=[], coefs=[],
                  std_errs=[], t_scores=[],p_values=[], alpha=None):
    """
    Generate a summary table of estimation results using Statsmodels SimpleTable. Still a
    work in progress.

    SimpleTable is maddening to work with, so it would be nice to find an alternative. It
    would need to support pretty-printing of formatted tables to plaintext and ideally
    also to HTML and Latex.

    At first it looked like we could use Statsmodels's summary table generator directly
    (iolib.summary.Summary), but this requires a Statsmodels results object as input and
    doesn't document which properties are pulled from it. PyLogit reverse engineered this
    for use in get_statsmodels_summary() -- so it's possible, but could be hard to
    maintain in the long run.

    We can't use PyLogit's summary table generator either. It requires a PyLogit
    model class as input, and we can't create one from results parameters. Oh well!

    """
    def fmt(value, format_str):
        # Custom numeric->string formatter that gracefully accepts null values
        return '' if value is None else format_str.format(value)

    if (title is None):
        title = "CHOICEMODELS ESTIMATION RESULTS"

    top_left = [['Dep. Var.:', dep_var],
                ['Model:', model_name],
                ['Method:', method],
                ['Date:', date],
                ['Time:', time],
                ['AIC:', fmt(aic, "{:,.3f}")],
                ['BIC:', fmt(bic, "{:,.3f}")]]

    top_right = [['No. Observations:', fmt(num_obs, "{:,}")],
                 ['Df Residuals:', fmt(df_resid, "{:,}")],
                 ['Df Model:', fmt(df_model, "{:,}")],
                 ['Pseudo R-squ.:', fmt(rho_squared, "{:.3f}")],
                 ['Pseudo R-bar-squ.:', fmt(rho_bar_squared, "{:.3f}")],
                 ['Log-Likelihood:', fmt(log_likelihood, "{:,.3f}")],
                 ['LL-Null:', fmt(null_log_likelihood, "{:,.3f}")]]

    # Zip into a single table (each side needs same number of entries)
    header_cells = [top_left[i] + top_right[i] for i in range(len(top_left))]

    # See end of statsmodels.iolib.table.py for formatting options
    header_fmt = dict(table_dec_below = '',
                      data_aligns = 'lrlr',
                      colwidths = 11,
                      colsep = '   ',
                      empty_cell = '')

    header = SimpleTable(header_cells, title=title, txt_fmt=header_fmt)

    col_labels = ['coef', 'std err', 'z', 'P>|z|', 'Conf. Int.']
    row_labels = x_names

    body_cells = [[fmt(coefs[i], "{:,.4f}"),
                   fmt(std_errs[i], "{:,.3f}"),
                   fmt(t_scores[i], "{:,.3f}"),
                   fmt(p_values[i], "{:,.3f}"),
                   '']  # conf int placeholder
                   for i in range(len(x_names))]

    body_fmt = dict(table_dec_below = '=',
                    header_align = 'r',
                    data_aligns = 'r',
                    colwidths = 7,
                    colsep = '   ')

    body = SimpleTable(body_cells,
                       headers = col_labels,
                       stubs = row_labels,
                       txt_fmt = body_fmt)

    # Ideally we'd want to append these into a single table, but I can't get it to work
    # without completely messing up the formatting..

    return (header, body)


"""
#############################
ORIGINAL FUNCTION DEFINITIONS
#############################

"""


logger = logging.getLogger(__name__)

# right now MNL can only estimate location choice models, where every equation
# is the same
# it might be better to use stats models for a non-location choice problem

# data should be column matrix of dimensions NUMVARS x (NUMALTS*NUMOBVS)
# beta is a row vector of dimensions 1 X NUMVARS


def mnl_probs(data, beta, numalts):
    logging.debug('start: calculate MNL probabilities')
    clamp = data.typ == 'numpy'
    utilities = beta.multiply(data)
    if numalts == 0:
        raise Exception("Number of alternatives is zero")
    utilities.reshape(numalts, utilities.size() // numalts)

    # https://stats.stackexchange.com/questions/304758/softmax-overflow
    utilities = utilities.subtract(utilities.max(0))
    exponentiated_utility = utilities.exp(inplace=True)
    if clamp:
        exponentiated_utility.inftoval(1e20)
        exponentiated_utility.clamptomin(1e-300)
    sum_exponentiated_utility = exponentiated_utility.sum(axis=0)
    probs = exponentiated_utility.divide_by_row(
        sum_exponentiated_utility, inplace=True)
    if clamp:
        probs.nantoval(1e-300)
        probs.clamptomin(1e-300)

    logging.debug('finish: calculate MNL probabilities')
    return probs


def get_hessian(derivative):
    return np.linalg.inv(np.dot(derivative, np.transpose(derivative)))


def get_standard_error(hessian):
    return np.sqrt(np.diagonal(hessian))

# data should be column matrix of dimensions NUMVARS x (NUMALTS*NUMOBVS)
# beta is a row vector of dimensions 1 X NUMVARS


def mnl_loglik(beta, data, chosen, numalts, weights=None, lcgrad=False,
               stderr=0):
    logger.debug('start: calculate MNL log-likelihood')
    numvars = beta.size
    numobs = data.size() // numvars // numalts

    beta = np.reshape(beta, (1, beta.size))
    beta = PMAT(beta, data.typ)

    probs = mnl_probs(data, beta, numalts)

    # lcgrad is the special gradient for the latent class membership model
    if lcgrad:
        assert weights
        gradmat = weights.subtract(probs).reshape(probs.size(), 1)
        gradarr = data.multiply(gradmat)
    else:
        if not weights:
            gradmat = chosen.subtract(probs).reshape(probs.size(), 1)
        else:
            gradmat = chosen.subtract(probs).multiply_by_row(
                weights).reshape(probs.size(), 1)
        gradarr = data.multiply(gradmat)

    if stderr:
        gradmat = data.multiply_by_row(gradmat.reshape(1, gradmat.size()))
        gradmat.reshape(numvars, numalts * numobs)
        return get_standard_error(get_hessian(gradmat.get_mat()))

    chosen.reshape(numalts, numobs)
    if weights is not None:
        if probs.shape() == weights.shape():
            loglik = ((probs.log(inplace=True)
                       .element_multiply(weights, inplace=True)
                       .element_multiply(chosen, inplace=True))
                      .sum(axis=1).sum(axis=0))
        else:
            loglik = ((probs.log(inplace=True)
                       .multiply_by_row(weights, inplace=True)
                       .element_multiply(chosen, inplace=True))
                      .sum(axis=1).sum(axis=0))
    else:
        loglik = (probs.log(inplace=True).element_multiply(
            chosen, inplace=True)).sum(axis=1).sum(axis=0)

    if loglik.typ == 'numpy':
        loglik, gradarr = loglik.get_mat(), gradarr.get_mat().flatten()
    else:
        loglik = loglik.get_mat()[0, 0]
        gradarr = np.reshape(gradarr.get_mat(), (1, gradarr.size()))[0]

    logger.debug('finish: calculate MNL log-likelihood')
    return -1 * loglik, -1 * gradarr


def mnl_estimate(data, chosen, numalts, GPU=False, coeffrange=(-1000, 1000),
                 weights=None, lcgrad=False, beta=None):
    """
    Calculate coefficients of the MNL model.

    Parameters
    ----------
    data : 2D array
        The data are expected to be in "long" form where each row is for
        one alternative. Alternatives are in groups of `numalts` rows per
        choosers. Alternatives must be in the same order for each chooser.
    chosen : 2D array
        This boolean array has a row for each chooser and a column for each
        alternative. The column ordering for alternatives is expected to be
        the same as their row ordering in the `data` array.
        A one (True) indicates which alternative each chooser has chosen.
    numalts : int
        The number of alternatives.
    GPU : bool, optional
    coeffrange : tuple of floats, optional
        Limits of (min, max) to which coefficients are clipped.
    weights : ndarray, optional
    lcgrad : bool, optional
    beta : 1D array, optional
        Any initial guess for the coefficients.

    Returns
    -------
    log_likelihood : dict
        Dictionary of log-likelihood values describing the quality of
        the model fit. The  following keys is entered into `log_likelihood`.
        
        aic : float
            Akaike information criterion for an estimated model. 
            aic =  -2 * log_likelihood + 2 * num_estimated_parameters 
        
        bic : float
            Bayesian information criterion for an estimated model. 
            bic = -2 * log_likelihood + log(num_observations) * num_parameters
        
        num_obs : int
            Number of observations in the model.

        df_model : int
            The number of parameters estimated in this model.

        df_resid : int
            The number of observations minus the number of estimated parameters.

        llnull : float
            Value of the constant-only loglikelihood
        
        ll : float
            Value of the loglikelihood

        rho_squared  : float
            McFadden's pseudo-R-squared. 
            rho_squared = 1.0 - final_log_likelihood / null_log_likelihood

        rho_bar_squared : float
            rho_bar_squared = 1.0 - ((final_log_likelihood - num_est_parameters) /
                             null_log_likelihood)


    fit_parameters : pandas.DataFrame
        Table of fit parameters with columns 'Coefficient', 'Std. Error',
        'T-Score', and 'P-Values. Each row corresponds to a column in `data` and are given
        in the same order as in `data`.

    See Also
    --------
    scipy.optimize.fmin_l_bfgs_b : The optimization routine used.

    """
    logger.debug(
        'start: MNL fit with len(data)={} and numalts={}'.format(
            len(data), numalts))
    atype = 'numpy' if not GPU else 'cuda'

    numvars = data.shape[1]
    numobs = data.shape[0] // numalts

    if chosen is None:
        chosen = np.ones((numobs, numalts))  # used for latent classes

    data = np.transpose(data)
    chosen = np.transpose(chosen)

    data, chosen = PMAT(data, atype), PMAT(chosen, atype)
    if weights is not None:
        weights = PMAT(np.transpose(weights), atype)

    if beta is None:
        beta = np.zeros(numvars)
    bounds = [coeffrange] * numvars

    # scipy optimization for MNL fit
    logger.debug('start: scipy optimization for MNL fit')
    args = (data, chosen, numalts, weights, lcgrad)
    bfgs_result = scipy.optimize.fmin_l_bfgs_b(mnl_loglik,
                                               beta,
                                               args=args,
                                               fprime=None,
                                               factr=10,
                                               approx_grad=False,
                                               bounds=bounds)
    logger.debug('finish: scipy optimization for MNL fit')

    if bfgs_result[2]['warnflag'] > 0:
        logger.warning("mnl did not converge correctly: %s",  bfgs_result)
    
    beta = bfgs_result[0]
    stderr = mnl_loglik(
        beta, data, chosen, numalts, weights, stderr=1, lcgrad=lcgrad)

    l0beta = np.zeros(numvars)
    l0 = -1 * mnl_loglik(l0beta, *args)[0]
    l1 = -1 * mnl_loglik(beta, *args)[0]
    ll = float(l1[0][0])
    ll_null = float(l0[0][0])
    rho_squared = 1.0 - ll / ll_null
    rho_bar_squared = 1.0 - ((ll - len(beta)) /ll_null)
    num_obs = numobs
    df_resid = numobs - len(beta)
    p_values = 2 * scipy.stats.norm.sf(np.abs(beta / stderr))
    bic = -2 * ll + np.log(numobs) * len(beta)
    aic = -2 * ll + 2 * len(beta)

    log_likelihood = {
        'null': float(l0[0][0]),
        'convergence': float(l1[0][0]),
        'ratio': float((1 - (l1 / l0))[0][0]),
        'rho_bar_squared': rho_bar_squared,
        'rho_squared': rho_squared,
        'df_model': len(beta),
        'df_resid': df_resid,
        'num_obs':numobs,
        'bic':bic,
        'aic':aic}

    fit_parameters = pd.DataFrame({
        'Coefficient': beta,
        'Std. Error': stderr,
        'T-Score': beta / stderr,
        'P-Values' : p_values })

    logger.debug('finish: MNL fit')
    return log_likelihood, fit_parameters



from setuptools import setup

# read README as the long description
with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='choicemodels',
    version='0.2.2',
    description='Tools for discrete choice estimation',
    long_description=long_description,
    author='UDST',
    url='https://github.com/udst/choicemodels',
    classifiers=[
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: BSD License'
    ],
    packages=['choicemodels', 'choicemodels.tools'],
    install_requires=[
        'numpy >= 1.14',
        'pandas >= 0.23',
        'patsy >= 0.5',
        'pylogit >= 0.2.2',
        'scipy >= 1.0',
        'statsmodels >= 0.8, <0.11; python_version <"3.6"',
        'statsmodels >= 0.8; python_version >="3.6"'
    ]
)
